from semantic_segmentation.pascal_voc_dataset import PascalVOCDataset  # NOQA
from image_retrieval.online_products_dataset import OnlineProductsDataset  # NOQA
from image_retrieval.online_products_dataset import get_online_products  # NOQA
